import os
from datetime import datetime
import logging

from badger.settings import init_settings
from badger.archive import BADGER_ARCHIVE_ROOT
from badger.errors import BadgerConfigError, BadgerLogbookError

logger = logging.getLogger(__name__)

# Check badger logbook root
config_singleton = init_settings()
BADGER_LOGBOOK_ROOT = config_singleton.read_value("BADGER_LOGBOOK_ROOT")
if BADGER_LOGBOOK_ROOT is None:
    raise BadgerConfigError("Please set the BADGER_LOGBOOK_ROOT env var!")
elif not os.path.exists(BADGER_LOGBOOK_ROOT):
    os.makedirs(BADGER_LOGBOOK_ROOT)
    logger.info(f"Badger logbook root {BADGER_LOGBOOK_ROOT} created")


def send_to_logbook(routine, widget=None):
    from xml.etree import ElementTree
    from re import sub

    log_text = ""
    routine_name = routine.name
    generator_name = routine.generator.name
    data_path = BADGER_ARCHIVE_ROOT
    obj_name = routine.vocs.objective_names[0]
    env_name = routine.environment.name

    idx_opt, obj_opt, _ = routine.vocs.select_best(routine.sorted_data, n=1)
    idx_opt = int(idx_opt[0])
    obj_opt = obj_opt[0]

    data = routine.data
    obj_start = data[obj_name].iloc[0]
    duration = data["timestamp"].iloc[-1] - data["timestamp"].iloc[0]
    n_point = len(data["timestamp"])
    if n_point > 0:
        log_text = f"Gain ({obj_name}): {round(obj_start, 4)} -> {round(obj_opt, 4)}\n"
    log_text += f"Time cost: {round(duration, 2)}s\n"
    log_text += f"Points requested: {n_point}\n"
    log_text += f"Optimal solution index: {idx_opt}\n"
    log_text += f"Routine name: {routine_name}\n"
    log_text += f"Environment name: {env_name}\n"
    log_text += f"Optimization algorithm: {generator_name}\n"
    log_text += f"Data location: {data_path}\n"
    try:
        log_text += f"Log location: {BADGER_LOGBOOK_ROOT}\n"
    except:
        pass

    # Generate the xml data
    curr_time = datetime.now()
    if os.name == "nt":
        timestr = curr_time.strftime("%Y-%m-%dT%H%M%S")
    else:
        timestr = curr_time.strftime("%Y-%m-%dT%H:%M:%S")
    log_entry = ElementTree.Element(None)
    severity = ElementTree.SubElement(log_entry, "severity")
    location = ElementTree.SubElement(log_entry, "location")
    keywords = ElementTree.SubElement(log_entry, "keywords")
    time = ElementTree.SubElement(log_entry, "time")
    isodate = ElementTree.SubElement(log_entry, "isodate")
    log_user = ElementTree.SubElement(log_entry, "author")
    category = ElementTree.SubElement(log_entry, "category")
    title = ElementTree.SubElement(log_entry, "title")
    metainfo = ElementTree.SubElement(log_entry, "metainfo")
    imageFile = ElementTree.SubElement(log_entry, "link")
    imageFile.text = timestr + "-00.ps"
    thumbnail = ElementTree.SubElement(log_entry, "file")
    thumbnail.text = timestr + "-00.png"
    text = ElementTree.SubElement(log_entry, "text")
    log_entry.attrib["type"] = "LOGENTRY"
    category.text = "USERLOG"
    location.text = "not set"
    severity.text = "NONE"
    keywords.text = "none"
    time.text = curr_time.strftime("%H:%M:%S")
    isodate.text = curr_time.strftime("%Y-%m-%d")
    metainfo.text = timestr + "-00.xml"
    log_user.text = " "
    title.text = "Badger"
    text.text = log_text
    if text.text == "":
        text.text = " "  # If field is truly empty, ElementTree leaves off tag entirely which causes logbook parser to fail

    fileName = os.path.join(BADGER_LOGBOOK_ROOT, metainfo.text)
    fileName = fileName.rstrip(".xml")
    xmlFile = open(fileName + ".xml", "w")
    rawString = ElementTree.tostring(log_entry, "utf-8").decode("utf-8")
    parsedString = sub(r"(?=<[^/].*>)", "\n", rawString)
    xmlString = parsedString[1:]
    xmlFile.write(xmlString)
    xmlFile.write("\n")  # Close with newline so cron job parses correctly
    xmlFile.close()
    screenshot(widget, f"{fileName}.png")


def screenshot(widget, filename):
    """
    Takes a screenshot of the whole gui window, saves png and ps images to file
    """
    if widget is None:
        raise BadgerLogbookError("No widget to take screenshot on!")

    from PIL import Image

    pic = widget.grab()
    pic.save(filename)
    img = Image.open(filename)
    if img.mode in ("RGBA", "LA"):
        # https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html?highlight=eps#eps
        # logger.warning(f'Current figure mode "{img.mode}" cannot be directly saved to .ps and will be converted to "RGB" mode')
        img = img.convert("RGB")
    # img = img.scaled(400, 600)
    name = os.path.splitext(filename)[0]
    img.save(f"{name}.ps")



from PyQt5.QtWidgets import (
    QTextEdit,
    QHBoxLayout,
    QVBoxLayout,
    QCheckBox,
    QWidget,
    QMainWindow,
)
from badger.factory import get_generator_docs


class BadgerDocsWindow(QMainWindow):
    def __init__(self, parent, generator):
        super().__init__(parent=parent)

        self.generator = generator
        self.render_md = True
        self.docs = None

        self.init_ui()
        self.config_logic()
        self.load_docs()

    def init_ui(self):
        self.setWindowTitle(f"Docs for generator {self.generator}")
        self.resize(640, 640)

        doc_panel = QWidget(self)
        vbox = QVBoxLayout(doc_panel)

        # Toolbar
        toolbar = QWidget()
        hbox_tool = QHBoxLayout(toolbar)
        hbox_tool.setContentsMargins(0, 0, 0, 0)
        self.cb_md = cb_md = QCheckBox("Render as Markdown")
        cb_md.setChecked(True)
        hbox_tool.addStretch()
        hbox_tool.addWidget(cb_md)
        vbox.addWidget(toolbar)

        self.markdown_viewer = QTextEdit()
        self.markdown_viewer.setReadOnly(True)
        vbox.addWidget(self.markdown_viewer)

        self.setCentralWidget(doc_panel)

    def config_logic(self):
        self.cb_md.stateChanged.connect(self.switch_render_mode)

    def load_docs(self):
        try:
            self.docs = docs = get_generator_docs(self.generator)
        except Exception as e:
            self.docs = docs = str(e)

        if self.render_md:
            self.markdown_viewer.setMarkdown(docs)
        else:
            self.markdown_viewer.setText(docs)

    def update_docs(self, generator):
        self.generator = generator
        self.setWindowTitle(f"Docs for generator {generator}")
        self.load_docs()

    def switch_render_mode(self):
        self.render_md = self.cb_md.isChecked()
        self.load_docs()



from badger.settings import init_settings, mock_settings
from badger.actions.config import _config_path_var


def self_check(args):
    config = init_settings()
    # Reset Badger

    if args.reset:
        while True:
            _res = input(
                "Badger will be reset to the factory settings.\n"
                "Only the configurations and built-in plugins will be reset,\n"
                "your saved routines and data will not be touched.\n\n"
                "proceed (y/[n])? "
            )
            if _res == "y":
                break
            elif (not _res) or (_res == "n"):
                print("Reset cancelled.")
                return
            else:
                print(f"Invalid choice: {_res}")

        config.reset_settings()
        print("Badger has been reset to the factory settings.")
        return

    good = check_n_config_paths()
    if good:
        print("Badger is healthy!")


def check_n_config_paths(config_filepath=None):
    if config_filepath is not None:
        config = init_settings(config_filepath)
    else:
        config = init_settings()

    good = True
    all_bad = True  # if all config paths are empty, we'll suggest initialization
    issue_list = []

    for pname, pvalue in config._config.model_dump(by_alias=True).items():
        if config.read_value(pname) is None:
            good = False
            issue_list.append(pname)
        else:
            if pvalue["is_path"]:
                all_bad = False

    if not good:
        if all_bad:
            # Initial setup
            init = True
            while True:
                _res = input(
                    "It looks like this is your first time launching Badger. Would you like to initialize it now?\n"
                    "Proceed ([y]/n): "
                )
                if (not _res) or (_res == "y"):
                    init = True
                    break
                elif _res == "n":
                    init = False
                    break
                else:
                    print(f"Invalid choice: {_res}")

            if init:  # fill in the mock up settings so user can go immediately
                mock_settings()

                return True

        # Let the users deal with the issues
        while True:
            _res = input(
                f"\nFound {len(issue_list)} issue(s).\n"
                "Fix the issue(s) now ([y]/n)? "
            )
            if (not _res) or (_res == "y"):
                break
            elif _res == "n":
                return good
            else:
                print(f"Invalid choice: {_res}")

        fixed = True
        for pname in issue_list:
            try:
                print("")
                success = _config_path_var(pname)
                if not success:
                    fixed = False
                # TODO potential keyError here
            except KeyboardInterrupt:
                pass

        return fixed

    return good



import logging
from logging.config import dictConfig
from badger.utils import merge_params

LOG_CONFIG_DEFAULT = {
    "version": 1,
    "formatters": {
        "simple": {"format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"}
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "simple",
            "stream": "ext://sys.stdout",
        }
    },
    "root": {"level": "WARNING", "handlers": ["console"]},
}


# Do not call this function twice within one session
# it will screw up the logging system
def config_log(dict_config=None):
    _config = merge_params(LOG_CONFIG_DEFAULT, dict_config)
    dictConfig(_config)


def set_log_level(level):
    loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
    for logger in loggers:
        logger.setLevel(level)



from typing import TypedDict


class PlotOptions(TypedDict):
    n_grid: int
    n_grid_range: tuple[int, int]
    show_samples: bool
    show_prior_mean: bool
    show_feasibility: bool
    show_acq_func: bool


class ConfigurableOptions(TypedDict):
    plot_options: PlotOptions
    variable_1: int
    variable_2: int
    include_variable_2: bool



from multiprocessing import Event, Pipe, Process, Queue

from PyQt5.QtCore import pyqtSignal, QObject

from badger.core_subprocess import run_routine_subprocess


class CreateProcess(QObject):
    """
    This class is for creating processes that will be used to run the optimizations.

    Note:
        The new process will be started, but will be holding until the wait_event is set.
    """

    finished = pyqtSignal()
    subprocess_prepared = pyqtSignal(object)

    def create_subprocess(self) -> None:
        """
        Creates a new process and starts it.
        The process and the arguments passed to the process are then emitted on
        the subprocess_prepared signal.
        """
        self.stop_event = Event()
        self.pause_event = Event()
        self.data_queue = Queue()
        self.evaluate_queue = Pipe()
        self.wait_event = Event()

        new_process = Process(
            target=run_routine_subprocess,
            args=(
                self.data_queue,
                self.evaluate_queue,
                self.stop_event,
                self.pause_event,
                self.wait_event,
            ),
        )
        new_process.start()
        self.subprocess_prepared.emit(
            {
                "process": new_process,
                "stop_event": self.stop_event,
                "pause_event": self.pause_event,
                "data_queue": self.data_queue,
                "evaluate_queue": self.evaluate_queue,
                "wait_event": self.wait_event,
            }
        )
        self.finished.emit()



import multiprocessing
import time

import pandas as pd
import pytest


class TestCore:
    """
    This class is to provide unit test coverage for the core.py file.
    """

    @pytest.fixture
    def process_manager(self):
        from badger.gui.default.components.create_process import CreateProcess
        from badger.gui.default.components.process_manager import ProcessManager

        process_manager = ProcessManager()
        process_builder = CreateProcess()
        process_builder.subprocess_prepared.connect(process_manager.add_to_queue)
        process_builder.create_subprocess()

        yield process_manager

        process_manager.close_proccesses()

    @pytest.fixture(scope="session")
    def init_multiprocessing(self):
        multiprocessing.set_start_method("fork", force=True)

    @pytest.fixture(autouse=True, scope="function")
    def test_core_setup(self, *args, **kwargs) -> None:
        super(TestCore, self).__init__(*args, **kwargs)
        self.count = 0
        self.candidates = None
        self.points_eval_list = []
        self.candidates_list = []
        self.states = None

        data = {"x0": [0.5], "x1": [0.5], "x2": [0.5], "x3": [0.5]}
        data_eval_target = {
            "x0": [0.5],
            "x1": [0.5],
            "x2": [0.5],
            "x3": [0.5],
            "f": [5.0],
        }

        self.points = pd.DataFrame(data)

        self.points_eval_target = pd.DataFrame(data_eval_target)

    def test_run_routine_subprocess(
        self, process_manager, init_multiprocessing
    ) -> None:
        """
        A unit test to ensure the core functionality
        of run_routine_xopt is functioning as intended.
        """
        from badger.archive import save_tmp_run
        from badger.tests.utils import create_routine

        self.count = 0
        self.num_of_points = 3

        self.routine = create_routine()
        time.sleep(1)
        tmp_filename = save_tmp_run(self.routine)
        self.termination_condition = {
            "tc_idx": 0,
            "max_eval": 3,
        }
        process_with_args = process_manager.remove_from_queue()
        pause_event = process_with_args["pause_event"]
        data_queue = process_with_args["data_queue"]
        wait_event = process_with_args["wait_event"]
        routine_process = process_with_args["process"]
        data_queue = process_with_args["data_queue"]
        evaluate_queue = process_with_args["evaluate_queue"]

        arg_dict = {
            "routine_id": self.routine.id,
            "routine_filename": tmp_filename,
            "routine_name": self.routine.name,
            "variable_ranges": self.routine.vocs.variables,
            "initial_points": self.routine.initial_points,
            "evaluate": True,
            "termination_condition": self.termination_condition,
            "start_time": time.time(),
        }

        data_queue.put(arg_dict)
        wait_event.set()
        pause_event.set()

        time.sleep(0.20)
        routine_process.terminate()
        time.sleep(1)

        while evaluate_queue[1].poll():
            self.results = evaluate_queue[1].recv()

        # assert len(self.candidates_list) == self.count - 1

        assert len(self.results[0]) == self.num_of_points

        assert self.states is None

        """
        path = "./test.yaml"
        assert os.path.exists(path) is True
        os.remove("./test.yaml")
        """

    def test_convert_to_solution(self) -> None:
        pass

    def test_evaluate_points(self) -> None:
        """
        A unit test to ensure the core functionality of evaluate_points
        is functioning as intended.
        """
        from badger.tests.utils import create_routine

        routine = create_routine()

        assert routine.environment.get_variables(["x1", "x2"]) == {
            "x1": 0.5,
            "x2": 0.5,
        }
        evaluate_points_result = routine.evaluate_data(self.points)

        vocs_list = ["x0", "x1", "x2", "x3", "f"]
        assert (
            evaluate_points_result[vocs_list]
            .astype(float)
            .equals(self.points_eval_target.astype(float))
        )

    def test_run_turbo(self, process_manager, init_multiprocessing) -> None:
        """
        A unit test to ensure TuRBO can run in Badger.
        """
        return

        from badger.db import save_routine
        from badger.tests.utils import create_routine_turbo

        self.count = 0
        self.num_of_points = 3
        self.routine = create_routine_turbo()
        time.sleep(1)
        assert self.routine.generator.turbo_controller.best_value is None
        save_routine(self.routine)
        self.termination_condition = {
            "tc_idx": 0,
            "max_eval": 3,
        }
        process_with_args = process_manager.remove_from_queue()
        pause_event = process_with_args["pause_event"]
        data_queue = process_with_args["data_queue"]
        wait_event = process_with_args["wait_event"]
        routine_process = process_with_args["process"]
        data_queue = process_with_args["data_queue"]
        evaluate_queue = process_with_args["evaluate_queue"]

        arg_dict = {
            "routine_id": self.routine.id,
            "evaluate": True,
            "termination_condition": self.termination_condition,
            "start_time": time.time(),
        }

        data_queue.put(arg_dict)
        wait_event.set()
        pause_event.set()

        time.sleep(0.20)
        routine_process.terminate()
        time.sleep(1)

        while evaluate_queue[1].poll():
            self.results = evaluate_queue[1].recv()

        # assert len(self.candidates_list) == self.count - 1

        # assert len(self.results) == self.num_of_points

        assert self.states is None

        """
        path = "./test.yaml"
        assert os.path.exists(path) is True
        os.remove("./test.yaml")
        """



from badger import interface


class Interface(interface.Interface):
    name = "default"
    # If params not specified, it would be an empty dict

    # Private variables
    _states: dict

    def __init__(self, **data):
        super().__init__(**data)

        self._states = {}

    def get_values(self, channel_names):
        channel_outputs = {}

        for channel in channel_names:
            try:
                value = self._states[channel]
            except KeyError:
                self._states[channel] = value = 0

            channel_outputs[channel] = value

        return channel_outputs

    def set_values(self, channel_inputs):
        for channel, value in channel_inputs.items():
            self._states[channel] = value



from importlib import metadata
from badger.actions.doctor import check_n_config_paths
from badger.utils import yprint
from badger.log import set_log_level


def show_info(args):
    # Change log level for all existed loggers
    set_log_level(args.log)
    config_path = None

    if args.config_filepath:
        config_path = args.config_filepath

    if args.gui:
        if check_n_config_paths(args.config_filepath):
            from badger.gui.acr import launch_gui

            launch_gui(config_path)

        return

    if args.gui_acr:
        if check_n_config_paths(args.config_filepath):
            from badger.gui.acr import launch_gui

            launch_gui(config_path)

        return

    if not check_n_config_paths():
        return

    from badger.settings import init_settings

    config_singleton = init_settings()
    BADGER_PLUGIN_ROOT = config_singleton.read_value("BADGER_PLUGIN_ROOT")
    BADGER_TEMPLATE_ROOT = config_singleton.read_value("BADGER_TEMPLATE_ROOT")
    BADGER_LOGBOOK_ROOT = config_singleton.read_value("BADGER_LOGBOOK_ROOT")
    BADGER_ARCHIVE_ROOT = config_singleton.read_value("BADGER_ARCHIVE_ROOT")

    info = {
        "name": "Badger the optimizer",
        "version": metadata.version("badger-opt"),
        "xopt version": metadata.version("xopt"),
        "plugin root": BADGER_PLUGIN_ROOT,
        "template root": BADGER_TEMPLATE_ROOT,
        "logbook root": BADGER_LOGBOOK_ROOT,
        "archive root": BADGER_ARCHIVE_ROOT,
        # 'plugin installation url': read_value('BADGER_PLUGINS_URL')
    }

    # print()  # put one blank line before the printout
    yprint(info)



import shutil
from os.path import exists
import logging

logger = logging.getLogger(__name__)


def plugin_remove(args):
    print(
        "This command is currently disabled.\n"
        "Please refer to the Badger documentation for plugin management.\n\n"
        "Badger online documentation: https://xopt-org.github.io/Badger/"
    )
    return

    try:
        from badger.factory import BADGER_PLUGIN_ROOT
    except Exception as e:
        logger.error(e)
        return

    hist = {
        "generator": "generators",
        "env": "environments",
        "ext": "extensions",
        "intf": "interfaces",
    }

    if args.plugin_type is None or args.plugin_specific is None:
        print("Please specify further which plugin you wish to remove!")
        return
    try:
        full_word = hist[f"{args.plugin_type}"]
    except KeyError:
        logger.error(f"{args.plugin_type} is not an existing plugin type")
        return

    plugin_path = f"{BADGER_PLUGIN_ROOT}/{full_word}/{args.plugin_specific}"
    if not exists(plugin_path):
        print(f"The plugin {args.plugin_specific} already does not exist!")
        return
    try:
        shutil.rmtree(plugin_path)
    except OSError as e:
        print(f"Error: {plugin_path} : {e.strerror}")
    else:
        print(
            f"{args.plugin_specific} was removed successfully from {BADGER_PLUGIN_ROOT}/{full_word}"
        )
        print(
            f"NOTE: The plugin dependencies for {args.plugin_specific} have not been removed"
        )



from copy import deepcopy
import logging
import time
import traceback
from pandas import DataFrame
import multiprocessing as mp

from badger.archive import load_run
from badger.errors import BadgerRunTerminated, BadgerEnvObsError
from badger.logger import _get_default_logger
from badger.logger.event import Events
from badger.routine import Routine
from badger.archive import archive_run

from xopt.errors import FeasibilityError, XoptError

logger = logging.getLogger(__name__)


def convert_to_solution(result: DataFrame, routine: Routine):
    """
    This method is passed the latest evaluated solution and converts that to a printable format for the terminal.
    This method is for the GUI version of Badger.

    Parameters
    ----------
    result : DataFrame
    routine : Routine
    """
    vocs = routine.vocs
    try:
        best_idx, _, _ = vocs.select_best(routine.sorted_data, n=1)

        if best_idx.size > 0:
            if best_idx[0] != len(routine.data) - 1:
                is_optimal = False
            else:
                is_optimal = True
        else:  # no feasible solution
            is_optimal = False

    except NotImplementedError:
        is_optimal = False  # disable the optimal highlight for MO problems
    except FeasibilityError:  # no feasible data
        logger.info("no feasible solutions found")
        is_optimal = False

    vars = list(result[vocs.variable_names].to_numpy()[0])
    objs = list(result[vocs.objective_names].to_numpy()[0])
    cons = list(result[vocs.constraint_names].to_numpy()[0])
    stas = list(result[vocs.observable_names].to_numpy()[0])

    # TODO: This structure needs improvement
    solution = (
        vars,
        objs,
        cons,
        stas,
        is_optimal,
        vocs.variable_names,
        vocs.objective_names,
        vocs.constraint_names,
        vocs.observable_names,
    )

    return solution


def run_routine_subprocess(
    queue: mp.Queue,
    evaluate_queue: mp.Pipe,
    stop_process: mp.Event,
    pause_process: mp.Event,
    wait_event: mp.Event,
) -> None:
    """
    Run the provided routine object using Xopt. This method is run as a subproccess

    Parameters
    ----------
    queue: mp.Queue
    evaluate_queue: mp.Pipe
    stop_process: mp.Event
    pause_process: mp.Event
    wait_event: mp.Event
    """
    wait_event.wait()

    try:
        args = queue.get(timeout=1)
    except Exception as e:
        print(f"Error in subprocess: {type(e).__name__}, {str(e)}")

    # set required arguments
    try:
        routine = load_run(args["routine_filename"])

        # Let interfaces reset any global state
        routine.environment.reset_environment()

        # Patch env with override variable ranges
        if routine.vrange_hard_limit:
            routine.environment.variables.update(routine.vrange_hard_limit)

        # TODO: might need to consider the case where routine.data is None?
        if routine.data is not None:
            routine.data = routine.data.iloc[0:0]  # reset the data
    except Exception as e:
        error_title = f"{type(e).__name__}: {e}"
        error_traceback = traceback.format_exc()
        queue.put((error_title, error_traceback))
        raise e

    # TODO look into this bug with serializing of turbo. Fix might be needed in Xopt
    # Patch for converting dtype str to torch object
    try:
        dtype = routine.generator.turbo_controller.tkwargs["dtype"]
        routine.generator.turbo_controller.tkwargs["dtype"] = eval(dtype)
    except AttributeError:
        pass
    except KeyError:
        pass
    except TypeError:
        pass

    # Assign the initial points and bounds
    routine.vocs.variables = args["variable_ranges"]
    routine.initial_points = args["initial_points"]

    # set optional arguments
    evaluate = args.pop("evaluate", None)
    archive = args.pop("archive", False)
    termination_condition = args.pop("termination_condition", None)
    start_time = args.pop("start_time", None)
    verbose = args.pop("verbose", 2)
    testing = args.pop("testing", False)

    # setup variables of routine properties for code readablilty
    initial_points = routine.initial_points

    # Log the optimization progress in terminal
    opt_logger = _get_default_logger(verbose)

    # Optimization starts
    # This is used by the logger to print to the terminal.
    solution_meta = (
        None,
        None,
        None,
        None,
        None,
        routine.vocs.variable_names,
        routine.vocs.objective_names,
        routine.vocs.constraint_names,
        routine.vocs.observable_names,
    )
    opt_logger.update(Events.OPTIMIZATION_START, solution_meta)

    # evaluate initial points:
    # timeout logic will be handled in the specific environment
    try:
        # initial sampling
        for _, ele in initial_points.iterrows():
            result = routine.evaluate_data(ele.to_dict())
            solution = convert_to_solution(result, routine)
            opt_logger.update(Events.OPTIMIZATION_STEP, solution)
            if evaluate:
                time.sleep(0.1)  # give it some break tp catch up
                evaluate_queue[0].send((routine.data, routine.generator))

        # optimization loop
        while True:
            if stop_process.is_set():
                evaluate_queue[0].close()
                raise BadgerRunTerminated
            elif not pause_process.is_set():
                pause_process.wait()

            # Check if termination condition has been satisfied
            if termination_condition and start_time:
                tc_config = termination_condition
                idx = tc_config["tc_idx"]
                if idx == 0:
                    max_eval = tc_config["max_eval"]
                    if len(routine.data) >= max_eval:
                        raise BadgerRunTerminated
                elif idx == 1:
                    max_time = tc_config["max_time"]
                    dt = time.time() - start_time
                    if dt >= max_time:
                        raise BadgerRunTerminated

            # TODO give user a message that a solution is being worked on.

            # generate points to observe
            candidates = routine.generator.generate(1)[0]
            candidates = DataFrame(candidates, index=[0])

            # TODO timer off + new timer for evaluate
            # TODO solution being evaluated

            # External triggers
            if stop_process.is_set():
                evaluate_queue[0].close()
                raise BadgerRunTerminated
            elif not pause_process.is_set():
                pause_process.wait()

            # if still active evaluate the points and add to generator
            # check active_callback evaluate point
            result = routine.evaluate_data(candidates)
            solution = convert_to_solution(result, routine)
            opt_logger.update(Events.OPTIMIZATION_STEP, solution)

            generator_copy = deepcopy(routine.generator)

            if evaluate:
                evaluate_queue[0].send((routine.data, generator_copy))

            # archive Xopt state after each step
            if archive:
                if not testing:
                    archive_run(routine)

    except BadgerRunTerminated:
        opt_logger.update(Events.OPTIMIZATION_END, solution_meta)
        evaluate_queue[0].close()
    except XoptError as e:
        opt_logger.update(Events.OPTIMIZATION_END, solution_meta)
        error_title = "BadgerEnvObsError: There was an error getting observables from the environment. See the traceback for more details."
        queue.put((error_title, traceback.format_exc()))
        evaluate_queue[0].close()
        raise BadgerEnvObsError(e)
    except Exception as e:
        opt_logger.update(Events.OPTIMIZATION_END, solution_meta)
        error_title = f"{type(e).__name__}: {e}"
        error_traceback = traceback.format_exc()
        queue.put((error_title, error_traceback))
        evaluate_queue[0].close()
        raise e



from badger.settings import init_settings
import os
import logging

from badger.utils import yprint, convert_str_to_value

logger = logging.getLogger(__name__)


def config_settings(args):
    config_singleton = init_settings()
    key = args.key

    if key is None:
        yprint(config_singleton.list_path_settings())
        return

    try:
        try:
            return _config_path_var(key)
        except KeyError:
            return _config_core_var(key)
    except KeyError:
        pass
    except IndexError:
        pass
    except KeyboardInterrupt:
        return

    logger.error(f"{key} is not a valid Badger config key!")


def _config_path_var(var_name):
    config_singleton = init_settings()

    is_path = config_singleton.read_is_path(var_name)

    if not is_path:
        raise KeyError

    display_name = config_singleton.read_display_name(var_name)
    desc = config_singleton.read_description(var_name)

    print(f"\n=== Configure {display_name} ===")
    print(f"*** {desc} ***\n")
    while True:
        res = input(
            f"Please type in the path to the Badger {display_name} folder (S to skip, R to reset): \n"
        )
        if res == "S":
            break
        if res == "R":
            _res = input(
                f"The current value {config_singleton.read_value(var_name)} will be reset, proceed (y/[n])? "
            )
            if _res == "y":
                break
            elif (not _res) or (_res == "n"):
                print("")
                continue
            else:
                print(f"Invalid choice: {_res}")

        res = os.path.abspath(os.path.expanduser(res))
        if os.path.isdir(res):
            _res = input(f"Your choice is {res}, proceed ([y]/n)? ")
            if _res == "n":
                print("")
                continue
            elif (not _res) or (_res == "y"):
                break
            else:
                print(f"Invalid choice: {_res}")
        else:
            _res = input(f"{res} does not exist, do you want to create it ([y]/n)? ")
            if _res == "n":
                print("")
                continue
            elif (not _res) or (_res == "y"):
                os.makedirs(res)
                print(f"Directory {res} has been created")
                break
            else:
                print(f"Invalid choice: {_res}")

    if res == "R":
        config_singleton.write_value(var_name, None)
        print(f"You reset the Badger {display_name} folder setting")
        return False  # fault value
    elif res != "S":
        config_singleton.write_value(var_name, res)
        print(f"You set the Badger {display_name} folder to {res}")
        return True  # success value
    else:
        print(f"You skipped the Badger {display_name} folder setting")
        return False


def _config_core_var(var_name):
    config_singleton = init_settings()

    display_name = config_singleton.read_display_name(var_name)
    desc = config_singleton.read_description(var_name)
    default = config_singleton.read_value(var_name)

    print(f"\n=== Configure {display_name} ===")
    print(f"*** {desc} ***\n")
    while True:
        res = input(
            f"Please type in the new value for {display_name} (S to skip, R to reset): \n"
        )
        if res == "S":
            break
        if res == "R":
            _res = input(
                f"The current value {config_singleton.read_value(var_name)} will be reset to {default}, proceed (y/[n])? "
            )
            if _res == "y":
                break
            elif (not _res) or (_res == "n"):
                print("")
                continue
            else:
                print(f"Invalid choice: {_res}")
        else:
            break

    if res == "R":
        config_singleton.write_value(var_name, default)
        print(f"You reset the {display_name} setting")
    elif res != "S":
        config_singleton.write_value(var_name, convert_str_to_value(res))
        print(f"You set {display_name} to {res}")



import os
import pytest
import pandas as pd
from badger.errors import BadgerRunTerminated


class TestCore:
    """
    This class is to provide unit test coverage for the core.py file.
    """

    @pytest.fixture(autouse=True, scope="function")
    def test_core_setup(self, *args, **kwargs) -> None:
        super(TestCore, self).__init__(*args, **kwargs)
        self.count = 0
        self.candidates = None
        self.points_eval_list = []
        self.candidates_list = []
        self.states = None

        data = {"x0": [0.5], "x1": [0.5], "x2": [0.5], "x3": [0.5]}
        data_eval_target = {
            "x0": [0.5],
            "x1": [0.5],
            "x2": [0.5],
            "x3": [0.5],
            "f": [5.0],
        }

        self.points = pd.DataFrame(data)

        self.points_eval_target = pd.DataFrame(data_eval_target)

    def mock_active_callback(self) -> int:
        """
        A mock active callback method to test
        if active callback is functioning properly.
        """
        if self.count >= 5:
            return 2
        else:
            return 0

    def mock_generate_callback(self, candidates: pd.DataFrame) -> None:
        """
        A mock generate callback method to test
        if generate callback is functioning properly.
        """
        self.candidates_list.append(candidates)

    def mock_evaluate_callback(self, points_eval: pd.DataFrame) -> None:
        """
        A mock evaluate callback method to test
        if evaluate callback is functioning properly.
        """
        self.points_eval_list.append(points_eval)
        self.count += 1

    def mock_states_callback(self, states: dict) -> None:
        """
        A mock states callback method to test
        if states callback is functioning properly.
        """
        self.states = states

    def mock_dump_file_callback(self) -> str:
        """
        A mock dump file callback method to test
        if dump file callback is functioning properly.
        """
        return "test.yaml"

    def test_run_routine(self) -> None:
        """
        A unit test to ensure the core functionality
        of run_routine_xopt is functioning as intended.
        """
        from badger.core import run_routine
        from badger.tests.utils import create_routine

        routine = create_routine()

        self.count = 0

        with pytest.raises(BadgerRunTerminated):
            run_routine(
                routine,
                self.mock_active_callback,
                self.mock_generate_callback,
                self.mock_evaluate_callback,
                self.mock_states_callback,
                self.mock_dump_file_callback,
            )

        assert len(self.candidates_list) == self.count - 1
        assert len(self.points_eval_list) == self.count

        assert self.states == {}

        path = "./test.yaml"
        assert os.path.exists(path) is True
        os.remove("./test.yaml")

    def test_evaluate_points(self) -> None:
        """
        A unit test to ensure the core functionality of evaluate_points
        is functioning as intended.
        """
        from badger.tests.utils import create_routine

        routine = create_routine()

        assert routine.environment.get_variables(["x1", "x2"]) == {
            "x1": 0.5,
            "x2": 0.5,
        }
        evaluate_points_result = routine.evaluate_data(self.points)

        vocs_list = ["x0", "x1", "x2", "x3", "f"]
        assert (
            evaluate_points_result[vocs_list]
            .astype(float)
            .equals(self.points_eval_target.astype(float))
        )

    def test_run_turbo(self) -> None:
        """
        A unit test to ensure TuRBO can run in Badger.
        """
        from badger.core import run_routine
        from badger.tests.utils import create_routine_turbo

        routine = create_routine_turbo()
        assert routine.generator.turbo_controller.best_value is None

        self.count = 0

        with pytest.raises(BadgerRunTerminated):
            run_routine(
                routine,
                self.mock_active_callback,
                self.mock_generate_callback,
                self.mock_evaluate_callback,
                self.mock_states_callback,
                dump_file_callback=None,
            )

        assert len(self.candidates_list) == self.count - 1
        assert len(self.points_eval_list) == self.count



import os
import shutil
import pytest


@pytest.fixture(autouse=True)
def suppress_popups(mocker):
    mocker.patch(
        "badger.gui.default.windows.expandable_message_box.ExpandableMessageBox.exec_",
        return_value=None,
    )


@pytest.fixture(scope="module", autouse=True)
def config_test_settings(
    mock_plugin_root, mock_template_root, mock_logbook_root, mock_archive_root
):
    from badger.settings import init_settings

    config_singleton = init_settings()

    # Store the old values
    old_root = config_singleton.read_value("BADGER_PLUGIN_ROOT")
    old_template = config_singleton.read_value("BADGER_TEMPLATE_ROOT")
    old_logbook = config_singleton.read_value("BADGER_LOGBOOK_ROOT")
    old_archived = config_singleton.read_value("BADGER_ARCHIVE_ROOT")

    # Assign values for test
    config_singleton.write_value("BADGER_PLUGIN_ROOT", mock_plugin_root)
    config_singleton.write_value("BADGER_TEMPLATE_ROOT", mock_template_root)
    config_singleton.write_value("BADGER_LOGBOOK_ROOT", mock_logbook_root)
    config_singleton.write_value("BADGER_ARCHIVE_ROOT", mock_archive_root)

    yield

    # Restoring the original settings
    config_singleton.write_value("BADGER_PLUGIN_ROOT", old_root)
    config_singleton.write_value("BADGER_TEMPLATE_ROOT", old_template)
    config_singleton.write_value("BADGER_LOGBOOK_ROOT", old_logbook)
    config_singleton.write_value("BADGER_ARCHIVE_ROOT", old_archived)


@pytest.fixture(scope="module", autouse=True)
def clean_up(mock_template_root, mock_logbook_root, mock_archive_root):
    # Clean before tests
    shutil.rmtree(mock_template_root, True)  # ignore errors
    shutil.rmtree(mock_logbook_root, True)
    shutil.rmtree(mock_archive_root, True)

    yield

    # Clean after tests
    shutil.rmtree(mock_template_root, True)
    shutil.rmtree(mock_logbook_root, True)
    shutil.rmtree(mock_archive_root, True)


@pytest.fixture(scope="module")
def mock_root(request):
    return os.path.join(request.fspath.dirname, "mock")


@pytest.fixture(scope="module")
def mock_plugin_root(mock_root):
    return os.path.join(mock_root, "plugins")


@pytest.fixture(scope="module")
def mock_template_root(mock_root):
    return os.path.join(mock_root, "templates")


@pytest.fixture(scope="module")
def mock_logbook_root(mock_root):
    return os.path.join(mock_root, "logbook")


@pytest.fixture(scope="module")
def mock_archive_root(mock_root):
    return os.path.join(mock_root, "archived")


@pytest.fixture(scope="module")
def mock_config_root(mock_root):
    return os.path.join(mock_root, "configs")



from typing import Optional, cast
from matplotlib import pyplot as plt
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from PyQt5.QtWidgets import QVBoxLayout, QWidget, QLayout
from xopt.generators.bayesian.visualize import (
    visualize_generator_model,
)
from xopt.generators.bayesian.bayesian_generator import BayesianGenerator

import time


import logging

logger = logging.getLogger(__name__)

plt.switch_backend("Qt5Agg")


class PlottingArea(QWidget):
    last_updated: Optional[float] = None

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)

        # Create a layout for the plot area without pre-filling it with a plot
        _layout = QVBoxLayout()
        self.setLayout(_layout)
        self.last_updated = time.time()

    def update_plot(
        self,
        generator: BayesianGenerator,
        variable_names: list[str],
        reference_point: dict[str, float],
        show_acquisition: bool,
        show_samples: bool,
        show_prior_mean: bool,
        show_feasibility: bool,
        n_grid: int,
        requires_rebuild: bool = False,
        interval: Optional[float] = 1000.0,  # Interval in milliseconds
    ):
        logger.debug("Updating plot in PlottingArea")

        debug_object = {
            "variable_names": variable_names,
            "reference_point": reference_point,
            "show_acquisition": show_acquisition,
            "show_samples": show_samples,
            "show_prior_mean": show_prior_mean,
            "show_feasibility": show_feasibility,
            "n_grid": n_grid,
            "requires_rebuild": requires_rebuild,
            "interval": interval,
        }

        # Check if the plot was updated recently
        if (
            self.last_updated is not None
            and interval is not None
            and not requires_rebuild
        ):
            logger.debug(f"Time since last update: {time.time() - self.last_updated}")

            time_diff = time.time() - self.last_updated

            # If the plot was updated less than 1 second ago, skip this update
            if time_diff < interval / 1000:
                logger.debug("Skipping update")
                return

        if generator.model is None:
            logger.debug("Model not found")
            return

        logger.debug(f"Arguments: {debug_object}")

        # Generate the new plot using visualize_generator_model
        fig, _ = cast(
            tuple[Figure, Axes],
            visualize_generator_model(
                generator,
                variable_names=variable_names,
                reference_point=reference_point,
                show_acquisition=show_acquisition,
                show_samples=show_samples,
                show_prior_mean=show_prior_mean,
                show_feasibility=show_feasibility,
                n_grid=n_grid,
            ),
        )

        # Adjust padding inside the figure
        fig.tight_layout(pad=1)  # Adds padding between plot elements

        layout = self.layout()

        if layout is not None:
            # Clear the existing layout (remove previous plot if any)
            self.clearLayout(layout)

            # Create a new figure and canvas
            canvas = FigureCanvas(fig)

            # Add the new canvas to the layout
            layout.addWidget(canvas)

            # Close the old figure to prevent memory leaks
            plt.close(fig)

        else:
            raise Exception("Layout never updated")

        # Ensure the layout is updated
        self.updateGeometry()
        self.adjustSize()

        # Update the last updated time
        self.last_updated = time.time()

    def clearLayout(self, layout: QLayout):
        while layout.count():
            child = layout.takeAt(0)
            if child is None:
                break

            widget = child.widget()
            if widget is None:
                break

            widget.deleteLater()



import os
import platform
import yaml
import shutil
from importlib import resources
from badger.utils import get_datadir
from pydantic import BaseModel, Field, ValidationError
from typing import Any, Dict, Optional, Union
from badger.errors import BadgerLoadConfigError


class Setting(BaseModel):
    """
    Setting model to store the configuration details.


    Attributes
    ----------
    display_name : str
        The display name of the setting.
    description : str
        A brief description of the setting.
    value : Optional[Union[str, int, bool, None]]
        The value of the setting which can be of different types.
    """

    display_name: str
    description: str
    value: Optional[Union[str, int, bool, None]] = Field(
        None, description="The value of the setting which can be of different types."
    )
    is_path: bool


class BadgerConfig(BaseModel):
    """
    BadgerConfig model to store file path, core configuration and GUI configuration details.

    Attributes
    ----------
    BADGER_PLUGIN_ROOT : Setting
        Setting for the plugin root directory.
    BADGER_TEMPLATE_ROOT : Setting
        Setting for the template root directory.
    BADGER_LOGBOOK_ROOT : Setting
        Setting for the logbook root directory.
    BADGER_ARCHIVE_ROOT : Setting
        Setting for the archive root directory.
    BADGER_DATA_DUMP_PERIOD : Setting
        Setting for the minimum time interval between data dumps (in seconds).
    BADGER_THEME : Setting
        Setting for the GUI theme.
    BADGER_ENABLE_ADVANCED : Setting
        Setting to enable advanced features in the GUI.
    """

    BADGER_PLUGIN_ROOT: Setting = Setting(
        display_name="plugin root",
        description="This setting (BADGER_PLUGIN_ROOT) tells Badger where to look for the plugins",
        value=None,
        is_path=True,
    )
    BADGER_TEMPLATE_ROOT: Setting = Setting(
        display_name="template root",
        description="This setting (BADGER_TEMPLATE_ROOT) tells Badger where to store the routine templates",
        value=None,
        is_path=True,
    )
    BADGER_LOGBOOK_ROOT: Setting = Setting(
        display_name="logbook root",
        description="This setting (BADGER_LOGBOOK_ROOT) tells Badger where to send the logs (GUI mode)",
        value=None,
        is_path=True,
    )
    BADGER_ARCHIVE_ROOT: Setting = Setting(
        display_name="archive root",
        description="This setting (BADGER_ARCHIVE_ROOT) tells Badger where to archive the historical optimization runs",
        value=None,
        is_path=True,
    )
    BADGER_DATA_DUMP_PERIOD: Setting = Setting(
        display_name="data dump period",
        description="Minimum time interval between data dumps, unit is second",
        value=1,
        is_path=False,
    )
    BADGER_THEME: Setting = Setting(
        display_name="theme",
        description="Theme for the Badger GUI",
        value="dark",
        is_path=False,
    )
    BADGER_ENABLE_ADVANCED: Setting = Setting(
        display_name="enable advanced features",
        description="Enable advanced features on the GUI",
        value=False,
        is_path=False,
    )
    AUTO_REFRESH: Setting = Setting(
        display_name="Auto-refresh",
        description="Permits each run to start from the initial points calculated based on the current values and the rules",
        value=False,
        is_path=False,
    )


class ConfigSingleton:
    _instance = None

    def __new__(cls, config_path: str = None, user_flag: bool = False):
        if cls._instance is None:
            cls._instance = super(ConfigSingleton, cls).__new__(cls)
            cls._instance.user_flag = user_flag
            cls._instance._config = cls.load_or_create_config(config_path)
            cls._instance.config_path = config_path
        return cls._instance

    @classmethod
    def load_or_create_config(cls, config_path: str) -> BadgerConfig:
        """
        Loads the config file from a given yaml file if it exists,
        otherwise creates an instance of BadgerConfig with default settings.

        Parameters
        ----------
        config_path: str
            Path to the user config file.

        Returns
        -------
        BadgerConfig
            An instance of BadgerConfig populated with the data from the config file,
            or with default settings if the file does not exist.
        """
        if os.path.exists(config_path):
            with open(config_path, "r") as config_file:
                config_data = yaml.safe_load(config_file)

            # Convert each entry in config_data to an instance of Setting
            for key, value in config_data.items():
                if isinstance(value, dict) and "value" in value:
                    # Convert to Setting instance
                    config_data[key] = Setting(
                        display_name=value.get("display_name", key),
                        description=value.get(
                            "description",
                            f"Setting for {key.replace('_', ' ').lower()}",
                        ),
                        value=value["value"],
                        is_path=value.get("is_path", key),
                    )
                else:
                    # If it's a direct value, wrap it in a Setting
                    config_data[key] = Setting(
                        display_name=key,
                        description=f"Setting for {key.replace('_', ' ').lower()}",
                        value=value,
                        is_path=False,
                    )

            try:
                return BadgerConfig(**config_data)
            except ValidationError as e:
                print(f"Error validating config file: {e}")
                raise
        else:
            if cls._instance.user_flag:
                err_msg = f"Error loading config {config_path}: invalid path."
                raise BadgerLoadConfigError(err_msg)

            return BadgerConfig()

    @property
    def config(self) -> BadgerConfig:
        return self._config

    def update_and_save_config(self, updates: Dict[str, Any]) -> None:
        """Saves changes to the config file.

        Parameters
        ----------
        updates : dict of {str: Any}
            A dictionary containing the dot-separated keys and new values to update in the YAML file.
        """
        config_data = self._config.model_dump(by_alias=True)

        # Apply updates to the config data using dot-separated keys
        for dot_key, value in updates.items():
            self._update_config_by_dot_key(config_data, dot_key, value)

        # Save updated config to file
        with open(self.config_path, "w") as file:
            yaml.dump(config_data, file, default_flow_style=False)

        self._config = BadgerConfig(**config_data)

    def _update_config_by_dot_key(
        self, config_data: Dict[str, Any], dot_key: str, value: Any
    ) -> None:
        """Update the config data with the provided value using dot-separated keys."""
        keys = dot_key.split(":")
        d = config_data
        for key in keys[:-1]:
            d = d.setdefault(key, {})
        last_key = keys[-1]
        if isinstance(d.get(last_key), dict) and "value" in d[last_key]:
            d[last_key]["value"] = value
        else:
            d[last_key] = value

    def list_settings(self) -> Dict[str, Any]:
        """List all the settings in Badger

        Returns
        -------
        result: Dict
            A dictionary containing the settings. Keys in the dict are fields of the
            settings, the value for each key is the current value for that setting.
        """
        return self._config.model_dump(by_alias=True)

    def list_path_settings(self) -> Dict[str, Any]:
        """List all the path-related settings in Badger

        Returns
        -------
        result: Dict
            A dictionary containing the path-related settings.
            Keys in the dict are fields of the settings,
            the value for each key is the current value for that setting.
        """
        path_dict = {}

        all_settings = self._config.model_dump(by_alias=True)
        for k, v in all_settings.items():
            if not v["is_path"]:
                continue

            # Skip the BADGER_DB_ROOT setting
            if k == "BADGER_DB_ROOT":
                continue

            path_dict[k] = v["value"]

        return path_dict

    def read_value(self, key: str, return_value_field: bool = True) -> Any:
        """
        Searches for the key in all sections of the configuration.

        Parameters
        ----------
        key: str
            The key to search for.
        return_value_field: bool
            If True, returns the 'value' field of the setting; otherwise, returns the entire setting.

        Returns
        -------
        Any
            The value associated with the provided key.

        Raises
        ------
        KeyError
            If the key is not found in the configuration.
        """
        data = self._config.model_dump(by_alias=True)
        if key in data:
            return data[key]["value"] if return_value_field else data[key]
        raise KeyError(f"Key '{key}' not found in the configuration.")

    def read_description(self, key: str, return_description_field: bool = True) -> Any:
        """
        Searches for the key in all sections of the configuration.

        Parameters
        ----------
        key: str
            The key to search for.
        return_description_field: bool
            If True, returns the 'description' field of the setting; otherwise, returns the entire setting.

        Returns
        -------
        str
            The description associated with the provided key.

        Raises
        ------
        KeyError
            If the key is not found in the configuration.
        """
        data = self._config.model_dump(by_alias=True)
        if key in data:
            return data[key]["description"] if return_description_field else data[key]
        raise KeyError(f"Key '{key}' not found in the configuration.")

    def read_display_name(
        self, key: str, return_display_name_field: bool = True
    ) -> Any:
        """
        Searches for the key in all sections of the configuration.

        Parameters
        ----------
        key: str
            The key to search for.
        return_display_name_field: bool
            If True, returns the 'display_name' field of the setting; otherwise, returns the entire setting.

        Returns
        -------
        str
            The display_name associated with the provided key.

        Raises
        ------
        KeyError
            If the key is not found in the configuration.
        """
        data = self._config.model_dump(by_alias=True)
        if key in data:
            return data[key]["display_name"] if return_display_name_field else data[key]
        raise KeyError(f"Key '{key}' not found in the configuration.")

    def read_is_path(self, key: str, return_is_path_field: bool = True) -> Any:
        """
        Searches for the key in all sections of the configuration.

        Parameters
        ----------
        key: str
            The key to search for.
        return_is_path_field: bool
            If True, returns the 'is_path' field of the setting; otherwise, returns the entire setting.

        Returns
        -------
        bool
            The is_path associated with the provided key.

        Raises
        ------
        KeyError
            If the key is not found in the configuration.
        """
        data = self._config.model_dump(by_alias=True)
        if key in data:
            return data[key]["is_path"] if return_is_path_field else data[key]
        raise KeyError(f"Key '{key}' not found in the configuration.")

    def write_value(self, key: str, value: Any) -> None:
        """A method for setting a new value to the config.

        Parameters
        ----------
        key: str
            The field that is being given a new value.
        value: Any
            The value that is being saved.
        """
        keys = key.split(".")
        updates = {}
        sub_dict = updates

        for key in keys[:-1]:
            sub_dict = sub_dict.setdefault(key, {})
        sub_dict[keys[-1]] = value

        self.update_and_save_config(updates)

    def reset_settings(self) -> None:
        """Resets all the settings to their default values."""
        default_config = BadgerConfig()
        self.update_and_save_config(default_config.model_dump(by_alias=True))
        print(
            f"All settings have been reset to their default values in {self.config_path}"
        )


def init_settings(config_arg: bool = None) -> ConfigSingleton:
    """
    Builds and returns an instance of the ConfigSingleton class.

    Parameters
    ----------
    config_arg: bool
        a path to a config file passed through the --config__filepath argument

    Returns
    -------
    config_singleton: ConfigSingleton
        an instance of the ConfigSingleton class
    """
    user_flag = False

    if config_arg is None:
        config_path = get_user_config_folder()
        file_name = "config.yaml"
        file_path = os.path.join(config_path, file_name)
    else:
        file_path = config_arg
        user_flag = True

    config_singleton = ConfigSingleton(file_path, user_flag)
    return config_singleton


def mock_settings():
    """A method for setting up mock settings"""
    config_singleton = init_settings()
    app_data_dir = get_datadir() / "Badger"
    os.makedirs(app_data_dir, exist_ok=True)

    # Configure the paths and put/refresh the mock plugins there if needed
    plugins_dir = str(app_data_dir / "plugins")
    os.makedirs(plugins_dir, exist_ok=True)
    config_singleton.write_value("BADGER_PLUGIN_ROOT", plugins_dir)
    built_in_plugins_dir = resources.files(__package__) / "built_in_plugins"
    shutil.copytree(built_in_plugins_dir, plugins_dir, dirs_exist_ok=True)

    logbook_dir = str(app_data_dir / "logbook")
    os.makedirs(logbook_dir, exist_ok=True)
    config_singleton.write_value("BADGER_LOGBOOK_ROOT", logbook_dir)

    archive_dir = str(app_data_dir / "archive")
    os.makedirs(archive_dir, exist_ok=True)
    config_singleton.write_value("BADGER_ARCHIVE_ROOT", archive_dir)

    templates_dir = str(app_data_dir / "templates")
    os.makedirs(templates_dir, exist_ok=True)
    config_singleton.write_value("BADGER_TEMPLATE_ROOT", templates_dir)

    # Set other settings to the default values
    for key in config_singleton.config.model_dump(by_alias=True).keys():
        config_singleton.write_value(
            key, config_singleton.config.model_dump(by_alias=True)[key]["value"]
        )


def get_user_config_folder() -> str:
    """
    Method for getting the path to the user specific folder on the current systems OS.

    Returns
    -------
    str
        The path to the user-specific configuration folder. This will be:
        - `%APPDATA%` or `%LOCALAPPDATA%` on Windows,
        - `~/Library/Application Support` on macOS,
        - `~/.config` on Linux.

    Raises
    ------
    OSError
        If the operating system is not supported.
    """
    system = platform.system()

    if system == "Windows":
        config_folder = os.getenv("APPDATA") or os.getenv("LOCALAPPDATA")
    elif system == "Darwin":
        config_folder = os.path.expanduser("~/Library/Application Support/Badger")
    elif system == "Linux":
        config_folder = os.path.expanduser("~/.config")
    else:
        raise OSError("Unsupported operating system")

    return config_folder



from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QWidget, QPushButton
from PyQt5.QtWidgets import QTextEdit, QStackedWidget
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QFont
from badger.gui.acr.components.routine_page import BadgerRoutinePage

from badger.routine import Routine


class BadgerRoutineEditor(QWidget):
    sig_saved = pyqtSignal()
    sig_canceled = pyqtSignal()
    sig_deleted = pyqtSignal()

    def __init__(self):
        super().__init__()

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(0, 0, 0, 0)

        # self.seperator = seperator = QFrame()
        # seperator.setFrameShape(QFrame.HLine)
        # seperator.setFrameShadow(QFrame.Sunken)
        # seperator.setLineWidth(0)
        # seperator.setMidLineWidth(0)
        # vbox.addWidget(seperator)

        # Routine stacks
        self.stacks = stacks = QStackedWidget()

        self.routine_edit = routine_edit = QTextEdit()
        routine_edit.setReadOnly(True)
        stacks.addWidget(routine_edit)

        # self.scroll_area = scroll_area = QScrollArea()
        self.routine_page = routine_page = BadgerRoutinePage()
        # scroll_area.setWidgetResizable(True)
        # scroll_area.setWidget(routine_page)
        stacks.addWidget(routine_page)

        stacks.setCurrentIndex(1)
        vbox.addWidget(stacks)

        # Action bar
        self.action_bar = action_bar = QWidget()
        # action_bar.hide()
        hbox_action = QHBoxLayout(action_bar)
        hbox_action.setContentsMargins(8, 0, 8, 0)

        cool_font = QFont()
        cool_font.setWeight(QFont.DemiBold)
        # cool_font.setPixelSize(16)

        # Only show when create new routine
        self.btn_cancel = btn_cancel = QPushButton("Cancel")
        btn_cancel.setFixedSize(64, 32)
        btn_cancel.setFont(cool_font)
        btn_cancel.hide()

        self.btn_save = btn_save = QPushButton("Save")
        btn_save.setFixedSize(128, 32)
        btn_save.setFont(cool_font)
        hbox_action.addWidget(btn_cancel)
        hbox_action.addStretch(1)
        hbox_action.addWidget(btn_save)
        btn_save.hide()
        # vbox.addWidget(action_bar)

    def config_logic(self):
        self.btn_cancel.clicked.connect(self.cancel_create_routine)

    def set_routine(self, routine: Routine, silent=False):
        try:
            self.routine_edit.setText(routine.yaml())
        except AttributeError:
            self.routine_edit.setText("")
        self.routine_page.refresh_ui(routine, silent=silent)

    def edit_routine(self):
        self.stacks.setCurrentIndex(1)

    def cancel_create_routine(self):
        self.sig_canceled.emit()

    def clear(self):
        self.routine_edit.clear()

    def switch_mode(self, mode):
        if mode == "regular":
            self.btn_cancel.hide()
        elif mode == "new routine":
            self.btn_cancel.show()



from badger import environment


class Environment(environment.Environment):
    name = "sphere_2d"
    variables = {
        "x0": [-1, 1],
        "x1": [-1, 1],
    }
    observables = ["f", "g"]

    _variables = {
        "x0": 0.5,
        "x1": 0.5,
    }
    _observations = {
        "f": 0.0,
        "g": 0.0,
    }

    def get_variables(self, variable_names):
        variable_outputs = {v: self._variables[v] for v in variable_names}

        return variable_outputs

    def set_variables(self, variable_inputs: dict[str, float]):
        for var, x in variable_inputs.items():
            self._variables[var] = x

        # Filling up the observations
        f = self._variables["x0"] ** 2 + self._variables["x1"] ** 2
        g = (self._variables["x0"] - 0.5) ** 2 + self._variables["x1"] ** 2

        self._observations["f"] = f
        self._observations["g"] = g

    def get_observables(self, observable_names):
        return {k: self._observations[k] for k in observable_names}
